import unittest
import pandas_ml_utils as pdu
import numpy as np
import pandas as pd


class TestTrainTestData(unittest.TestCase):

    def test_no_training_data(self):

        df = pd.DataFrame({"featureA": [1,2,3,4,5],
                           "featureB": [5,4,3,2,1],
                           "labelA": [1,2,3,4,5],
                           "labelB": [5,4,3,2,1]})

        x_train, x_test, y_train, y_test, _, _, _ = df.make_training_data(
            pdu.FeaturesAndLabels(["featureA", "featureB"], ["labelA"]),
            test_size=0)

        self.assertIsNone(x_test)
        self.assertIsNone(y_test)
        np.testing.assert_array_almost_equal(x_train, df[["featureA", "featureB"]].values)
        np.testing.assert_array_almost_equal(y_train, df["labelA"].values)

    def test_make_training_data(self):
        df = pd.DataFrame({"featureA": [1,2,3,4,5],
                           "featureB": [5,4,3,2,1],
                           "labelA": [1,2,3,4,5],
                           "labelB": [5,4,3,2,1]})

        x_train, x_test, y_train, y_test, _, _, _ = df.make_training_data(
            pdu.FeaturesAndLabels(["featureA", "featureB"], ["labelA"]),
            test_size=0.5)

        np.testing.assert_array_almost_equal(x_test, np.array([[2, 4], [5, 1], [3, 3]]))
        np.testing.assert_array_almost_equal(y_test, np.array([2, 5, 3]))

    def test_make_training_data_two_labels(self):
        df = pd.DataFrame({"featureA": [1,2,3,4,5],
                           "featureB": [5,4,3,2,1],
                           "labelA": [1,2,3,4,5],
                           "labelB": [5,4,3,2,1]})

        x_train, x_test, y_train, y_test, _, _, _  = df.make_training_data(
            pdu.FeaturesAndLabels(["featureA", "featureB"], ["labelA", "labelB"]),
            test_size=0.5)

        np.testing.assert_array_almost_equal(x_test, np.array([[2, 4], [5, 1], [3, 3]]))
        np.testing.assert_array_almost_equal(y_test, np.array([[2, 4], [5, 1], [3, 3]]))

    def test_make_rnn_training_data(self):
        df = pd.DataFrame({"featureA": [1,2,3,4,5],
                           "featureB": [5,4,3,2,1],
                           "labelA": [1,2,3,4,5],
                           "labelB": [5,4,3,2,1]})

        fl = pdu.FeaturesAndLabels(["featureA", "featureB"], ["labelA"], feature_lags=[0, 1])

        x_train, x_test, y_train, y_test, _, _, min = df.make_training_data(fl, test_size=0.5)

        # test whole shape and labels
        np.testing.assert_array_almost_equal(x_test, np.array([[[3, 3], [2, 4]], [[5, 1], [4, 2]]]))
        np.testing.assert_array_almost_equal(y_test, np.array([3, 5]))

        # all rows, all lags one feature -> feature[0] needs lag of -1
        np.testing.assert_array_almost_equal(x_test[:,:,0], np.array([[3, 2], [5, 4]]))
        # all rows, all lags one feature -> feature[1] needs lag of +1
        np.testing.assert_array_almost_equal(x_test[:,:,1], np.array([[3, 4], [1, 2]]))

        # test min
        self.assertEqual(min, 2)

        # test names
        self.assertListEqual(fl.get_feature_names().tolist(), [['featureA_0', 'featureB_0'],
                                                               ['featureA_1', 'featureB_1']])

    def test_make_single_lagged_training_data(self):
        df = pd.DataFrame({"featureA": [1,2,3,4,5],
                           "featureB": [5,4,3,2,1],
                           "labelA": [1,2,3,4,5],
                           "labelB": [5,4,3,2,1]})

        x_train, x_test, y_train, y_test, _, _, _ = df.make_training_data(
            pdu.FeaturesAndLabels(["featureA"], ["labelA"], feature_lags=[1, 2]),
            test_size=0.5)

        np.testing.assert_array_almost_equal(x_test, np.array([[[2], [1]], [[3], [2]]]))
        np.testing.assert_array_almost_equal(y_test, np.array([3, 4]))

    def test_make_smoothed_training_data(self):
        df = pd.DataFrame({"featureA": [1,2,3,4,5,6,7,8,9,10],
                           "featureB": [5,4,3,2,1,0,1,2,3,4],
                           "labelA": [1,2,3,4,5,6,7,8,9,10],
                           "labelB": [5,4,3,2,1,0,1,2,3,4]})

        x_train, x_test, y_train, y_test, _, _, min = df.make_training_data(
            pdu.FeaturesAndLabels(["featureA"], ["featureA"], feature_lags=[1, 2, 3, 4],
                                  lag_smoothing={2: lambda df: df[["featureA"]] * 2,
                                                 4: lambda df: df[["featureA"]] * 4}),
            test_size=0.5)

        np.testing.assert_array_almost_equal(x_train[-1], [[7], [12], [10], [16]])
        self.assertEqual(min, 5)

    def test_lag_smoothing_nan(self):
        # test lag smoothing using shift (introducing nan)
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "labelA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "labelB": [5, 4, 3, 2, 1, 0, 1, 2, 3, None]})

        from pandas_ml_utils.train_test_data import _make_features
        fl = pdu.FeaturesAndLabels(["featureA"], None, feature_lags=[0, 1],
                                   lag_smoothing={1: lambda df: df["featureA"].shift(2)})

        df, x = _make_features(df, fl)
        len_features = 10 - 1 - 2
        len_none_lables = 1

        self.assertEqual(len(df), len_features - len_none_lables)
        np.testing.assert_array_equal(fl.get_feature_names(), np.array( [['featureA_0'], ['featureA_1']]))
        self.assertAlmostEqual(df["featureA_1"].iloc[0], 1.0)
        self.assertAlmostEqual(df["featureA_1"].iloc[-1], 6.0)

    def test_reshape_rnn_as_ar(self):
        np.testing.assert_array_almost_equal(pdu.reshape_rnn_as_ar(np.array([[[1], [2]]], ndmin=3)),
                                             np.array([[1, 2]], ndmin=2))

        np.testing.assert_array_almost_equal(pdu.reshape_rnn_as_ar(np.array([[1, 2]], ndmin=2)),
                                             np.array([[1, 2]], ndmin=2))

    def test_hashable_features_and_labels(self):
        a = pdu.FeaturesAndLabels(["featureA"], ["featureA"], feature_lags=[1, 2, 3, 4],
                                  lag_smoothing={2: lambda df: df[["featureA"]] * 2,
                                                 4: lambda df: df[["featureA"]] * 4})

        b = pdu.FeaturesAndLabels(["featureA"], ["featureA"], feature_lags=[1, 2, 3, 4],
                                  lag_smoothing={2: lambda df: df[["featureA"]] * 2,
                                                 4: lambda df: df[["featureA"]] * 4})

        self.assertEqual(hash(a), hash(a))
        self.assertEqual(a, a)

        self.assertEqual(a.__id__(), b.__id__())
        self.assertEqual(a, b)
        self.assertEqual(hash(a), hash(b))

    def test_cache_feature_matrix(self):
        from pandas_ml_utils.train_test_data import make_training_data, _make_features_with_cache

        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "labelA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "labelB": [5, 4, 3, 2, 1, 0, 1, 2, 3, None]})

        ful = pdu.FeaturesAndLabels(["featureA"], ["featureA"], feature_lags=[1, 2, 3, 4],
                                    lag_smoothing={2: lambda df: df[["featureA"]] * 2,
                                                   4: lambda df: df[["featureA"]] * 4})

        x_train, x_test, y_train, y_test, index_train, index_test, min = make_training_data(df, ful, cache=True)
        x_train, x_test, y_train, y_test, index_train, index_test, min = make_training_data(df, ful, cache=True)

        cache_info = _make_features_with_cache.cache_info()
        print(cache_info)

        self.assertEqual(cache_info.misses, 1)  # miss the first call
        self.assertEqual(cache_info.hits, 1)    # fetch the first cached call

    def test_feature_scaling_3d(self):
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "featureC": [11, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "featureB": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                           "labelA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})

        fl = pdu.FeaturesAndLabels(["featureA", "featureB", "featureC"],
                                   ["labelA"],
                                   feature_lags=[1, 2],
                                   feature_rescaling={("featureA", "featureC"): (-1, 1)})

        x_train, x_test, y_train, y_test, _, _, _ = df.make_training_data(fl, test_size=0)

        print(x_train)
        self.assertEqual(x_train.argmax(), 5)
        self.assertEqual(x_train[0,1,2], 1)
        self.assertEqual(x_train[0,1,0], -1)
        np.testing.assert_array_almost_equal(x_train[0,:,1], df["featureB"][[1,0]])

    def test_feature_scaling_2d(self):
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "featureC": [11, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "featureB": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
                           "labelA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})

        fl = pdu.FeaturesAndLabels(["featureA", "featureB", "featureC"],
                                   ["labelA"],
                                   feature_rescaling={("featureA", "featureC"): (-1, 1)})

        x_train, x_test, y_train, y_test, _, _, _ = df.make_training_data(fl, test_size=0)

        print(x_train)
        np.testing.assert_array_almost_equal(x_train[0], np.array([-1, 0.1, 1]))

    def test_lagging(self):
        df = pd.DataFrame({"featureA": [0.5592344 , 0.60739384, 0.19994533, 0.56642537, 0.50965677,
                                        0.168989  , 0.94080671, 0.76651769, 0.8403563 , 0.4003567 ,
                                        0.24295908, 0.50706317, 0.66612371, 0.4020924 , 0.21776017,
                                        0.32559497, 0.12721287, 0.13904584, 0.65887554, 0.08830925],
                           "featureC": [0.43700002, 0.36804634, 0.37568437, 0.34575936, 0.45747071,
                                        0.49749949, 0.09991126, 0.64710179, 0.93479635, 0.40651901,
                                        0.22387503, 0.18458239, 0.0271485 , 0.68792433, 0.69638729,
                                        0.60715125, 0.52557556, 0.88319929, 0.34808869, 0.50250121],
                           "featureB": [0.6784767 , 0.184668  , 0.02049038, 0.92983967, 0.67628553,
                                        0.71373065, 0.30594832, 0.63038278, 0.78284284, 0.84566334,
                                        0.00558188, 0.15819783, 0.09404578, 0.34460875, 0.69161826,
                                        0.41633249, 0.51130681, 0.66703763, 0.74652599, 0.26560367],
                           "labelA": range(20)})

        fl = pdu.FeaturesAndLabels(["featureA", "featureB", "featureC"],
                                   ["labelA"],
                                   feature_lags=[0,1,2,3,4])

        x_train, x_test, y_train, y_test, _, _, _ = df.make_training_data(fl, test_size=0)

        self.assertEqual(len(x_train), len(df) - 4)
        np.testing.assert_array_equal(x_train[0,:,0], df["featureA"].values[[4,3,2,1,0]])
        np.testing.assert_array_equal(x_train[-1,:,0], df["featureA"].values[[-1, -2, -3, -4, -5]])
        np.testing.assert_array_equal(x_train[0,:,1], df["featureB"].values[[4,3,2,1,0]])
        np.testing.assert_array_equal(x_train[-1,:,1], df["featureB"].values[[-1, -2, -3, -4, -5]])
        np.testing.assert_array_equal(x_train[0,:,2], df["featureC"].values[[4,3,2,1,0]])
        np.testing.assert_array_equal(x_train[-1,:,2], df["featureC"].values[[-1, -2, -3, -4, -5]])

if __name__ == '__main__':
    unittest.main()



import inspect
import logging
from typing import List, Callable, Iterable, Dict, Type, Tuple, Union
from numbers import Number
import pandas as pd
import numpy as np

log = logging.getLogger(__name__)


class FeaturesAndLabels(object):
    """
    *FeaturesAndLabels* is the main object used to hold the context of your problem. Here you define which columns
    of your `DataFrame` is a feature, a label or a target. This class also provides some functionality to generate
    autoregressive features. By default lagging features results in an RNN shaped 3D array (in the format of keras
    RNN layers input format).
    """

    def __init__(self,
                 features: List[str],
                 labels: List[str],
                 label_type:Type = int,
                 targets: Union[List[str], Tuple[str, str], Dict[str, str], Dict[str, Tuple[str, List[str]]], Dict[str, Tuple[str, str]]] = None,
                 feature_lags: Iterable[int] = None,
                 feature_rescaling: Dict[Tuple[str], Tuple[int]] = None,
                 lag_smoothing: Dict[int, Callable[[pd.Series], pd.Series]] = None,
                 **kwargs):
        """
        :param features: a list of column names which are used as features for your model
        :param labels: as list of column names which are uses as labels for your model
        :param label_type: whether to treat a label as int, float, bool
        :param targets: targets are composed of a *target*, *loss*, *subset of labels* where any part can be None.
                        Technically at least one target is needed and defaults to ("target", -1, self.labels) but
                        multiple targets are possible. The concept around targets is like follows. Let's say you want
                        to classify whether a printer is jamming the next page or not. Halting and servicing the printer
                        costs 5'000 while a jam costs 15'000. Your target will be 0 or empty but your loss will be
                        -5000 for all your type II errors and -15'000 for all your type I errors in case of miss-
                        classification. Another example would be if you want to classify whether a stock price is
                        above (buy) the current price or not (do nothing). Your target is the today's price and your
                        loss is tomorrows price minus today's price.
                        Note: Not all :class:`.Model` support multiple targets
        :param feature_lags: an iterable of integers specifying the lags of an AR model i.e. [1] for AR(1)
                             if the un-lagged feature is needed as well provide also lag of 0 like range(1)
        :param feature_rescaling: this allows to rescale features.
                                  in a dict we can define a tuple of column names and a target range
        :param lag_smoothing: very long lags in an AR model can be a bit fuzzy, it is possible to smooth lags i.e. by
                              using moving averages. the key is the lag length at which a smoothing function starts to
                              be applied
        :param kwargs: maybe you want to pass some extra parameters to a model
        """
        self.features = features
        self.labels = labels
        self.label_type = label_type
        self.targets = targets
        self.feature_lags = [lag for lag in feature_lags] if feature_lags is not None else None
        self.feature_rescaling = feature_rescaling
        self.lag_smoothing = lag_smoothing
        self.len_feature_lags = sum(1 for _ in self.feature_lags) if self.feature_lags is not None else 1
        self.expanded_feature_length = len(features) * self.len_feature_lags if feature_lags is not None else len(features)
        self.min_required_samples = (max(feature_lags) + 1) if self.feature_lags is not None else 1
        self.kwargs = kwargs
        log.info(f'number of features, lags and total: {self.len_features()}')

    def shape(self):
        return self.get_feature_names().shape, (self.len_labels(), )

    def len_features(self):
        return len(self.features), self.len_feature_lags, self.expanded_feature_length

    def len_labels(self):
        return len(self.labels)

    def get_feature_names(self):
        if self.feature_lags is not None:
            return np.array([[f'{feat}_{lag}'
                              for feat in self.features]
                             for lag in self.feature_lags], ndmin=2)
        else:
            return np.array(self.features)

    def get_goals(self) -> Dict[str, Tuple[str, List[str]]]:
        # if we can return a dictionary of target -> (loss, labels) where loss will be a column or constant 1
        if isinstance(self.targets, Number) or isinstance(self.targets, str):
            # we have a target value but no loss
            return {self.targets: (None, self.labels)}
        elif self.targets is None or len(self.targets) <= 0:
            # we have got no target no loss an no labels
            return {None: (None, self.labels)}
        elif isinstance(self.targets, tuple):
            # we have a pair of target value and loss
            return {self.targets[0]: (self.targets[1], self.labels)}
        elif isinstance(self.targets, dict):
            sample_value = next(iter(self.targets.values()))
            if isinstance(sample_value, tuple) and len(sample_value) == 2 and isinstance(sample_value[1], list):
                # we have a dict of target values and tuple losses, labels
                return self.targets
            elif isinstance(sample_value, tuple) and len(sample_value) == 2 and isinstance(sample_value[1], str):
                # we have a dict of target values and tuple losses, labels
                return {k: (v[0], [v[1]]) for k, v in self.targets.items()}
            elif isinstance(sample_value, str) or isinstance(sample_value, Number):
                # we have a dict of target values and losses
                return {k: (v, self.labels) for k, v in self.targets.items()}
            else:
                raise ValueError("you need to provide a loss or a tuple[loss, list[str]")
        else:
            raise ValueError("you need to provide a traget column name oder a dictionary with target column name as key")

    def __getitem__(self, item):
        if isinstance(item, tuple) and len(item) == 2:
            return self.kwargs[item[0]] if item[0] in self.kwargs else item[1]
        else:
            return self.kwargs[item] if item in self.kwargs else None

    def __repr__(self):
        return f'FeaturesAndLabels({self.features},{self.labels},{self.targets},' \
               f'{self.feature_lags},{self.feature_rescaling}{self.lag_smoothing}) ' \
               f'#{len(self.features)} ' \
               f'features expand to {self.expanded_feature_length}'

    def __hash__(self):
        return hash(self.__id__())

    def __eq__(self, other):
        return self.__id__() == other.__id__()

    def __id__(self):
        import dill  # only import if really needed
        smoothers = ""

        if self.lag_smoothing is not None:
            smoothers = {feature: inspect.getsource(smoother) for feature, smoother in self.lag_smoothing.items()}

        return f'{self.features},{self.labels},{self.label_type},{self.targets},{dill.dumps(self.feature_lags)},{self.feature_rescaling},{smoothers}'

    def __str__(self):
        return self.__repr__()




from unittest import TestCase

from pandas_ml_utils.model.features_and_Labels import FeaturesAndLabels


class TestFeaturesAndLabels(TestCase):

    def test_id(self):
        """given"""
        fl1 = FeaturesAndLabels(["a", "b", "c"], ["d", "e"], targets={"b": None})
        fl2 = FeaturesAndLabels(["a", "b", "c"], ["d", "e"], targets={"b": None})
        fl3 = FeaturesAndLabels(["a", "b", "d"], ["d", "e"], targets={"b": None})

        """expect"""
        self.assertEqual(fl1.__id__(), fl2.__id__())
        self.assertEqual(hash(fl2), hash(fl2))
        self.assertNotEqual(hash(fl2), hash(fl3))

    def test_1d(self):
        """given"""
        fl = FeaturesAndLabels(["a", "b", "c"], ["d", "e"])

        """when"""
        shape = fl.shape()

        """then"""
        self.assertEqual(shape, ((3, ), (2, )))


    def test_2d(self):
        """given"""
        fl = FeaturesAndLabels(["a", "b", "c"], ["d", "e"], feature_lags=range(4))

        """when"""
        shape = fl.shape()

        """then"""
        # shape is ((timesteps, features), (labels, )
        self.assertEqual(shape, ((4, 3), (2, )))

    def test_goals(self):
        """given"""
        fl1 = FeaturesAndLabels(["a", "b", "c"], ["d", "e"], targets='a')
        fl2 = FeaturesAndLabels(["a", "b", "c"], ["d", "e"], targets=('a', 'b'))
        fl3 = FeaturesAndLabels(["a", "b", "c"], ["d", "e"], targets={'a': 'b'})
        fl4 = FeaturesAndLabels(["a", "b", "c"], ["d", "e"], targets={'a': ('b', 'd')})
        fl5 = FeaturesAndLabels(["a", "b", "c"], ["d", "e"], targets={'a': ('b', ['d'])})

        """when"""
        g1 = fl1.get_goals()
        g2 = fl2.get_goals()
        g3 = fl3.get_goals()
        g4 = fl4.get_goals()
        g5 = fl5.get_goals()

        """then"""
        self.assertEqual(g1, {'a': (None, ["d", "e"])})
        self.assertEqual(g2, {'a': ('b', ["d", "e"])})
        self.assertEqual(g3, {'a': ('b', ["d", "e"])})
        self.assertEqual(g4, {'a': ('b', ["d"])})
        self.assertEqual(g5, {'a': ('b', ["d"])})


from typing import Any

import pandas as pd
import os
from .models import Model
from .summary import Summary


class Fit(object):
    """
    After a model is fitted it gets embedded into this class along with some :class:`.Summary` statistics
    """

    def __init__(self,
                 model: Model,
                 training_summary: Summary,
                 test_summary: Summary,
                 trails: Any):
        self.model = model
        self.training_summary = training_summary
        self.test_summary = test_summary
        self._trails = trails

    def values(self):
        """
        :return: returns the fitted model, a :class:`.Summary` on the training data, a :class:`.Summary` on the test data
        """
        return self.model, self.training_summary, self.test_summary

    def trails(self):
        """
        In case of hyper parameter optimization a trails object as used by `Hyperopt <https://github.com/hyperopt/hyperopt/wiki/FMin>`_
        is available.

        :return: Trails object
        """
        if self._trails is not None:
            return pd.DataFrame(self._trails.results)\
                     .drop("parameter", axis=1)\
                     .join(pd.DataFrame([r['parameter'] for r in self._trails.results]))
        else:
            return None

    def save_model(self, filename: str):
        """
        Save the fitted model.

        :param filename: filename
        :return: None
        """
        self.model.save(filename)

    def _repr_html_(self):
        from mako.template import Template
        from mako.lookup import TemplateLookup

        template = Template(filename=f"{os.path.abspath(__file__)}.html", lookup=TemplateLookup(directories=['/']))
        return template.render(fit=self)


import inspect
import logging
import sys
from typing import List, Tuple, Callable, Iterable, Dict, Union
from matplotlib import pyplot as plt

import seaborn as sns
import numpy as np
import pandas as pd


log = logging.getLogger(__name__)


def plot_forecast_heatmap(df: pd.DataFrame,
                          multi_model: object,
                          parameter_as_column: str):

    # we need a data frame with the target values as row index and the forecast periods as columns
    # if we have one data frame (which is effectively one row) per "target" we just need to concatenete them
    # and join the same frames as new column for each forecast period
    pass


import logging
import sys
from typing import Tuple, Union, Dict

import numpy as np
import pandas as pd

from ..model.summary import Summary

log = logging.getLogger(__name__)


class RegressionSummary(Summary):

    def __init__(self, df: pd.DataFrame):
        self.df = df

    # TODO add some statistics and fix loss is always None at the moment
    #  mse, r2, ... pvalue?



import logging
from typing import Callable, Tuple, Dict

import numpy as np
import pandas as pd

from .summary import RegressionSummary
from ..model.fit import Fit
from ..model.fitter import _fit, _backtest, _predict
from ..model.models import Model
from ..error.functions import mse as _mse
from ..constants import *

log = logging.getLogger(__name__)


def fit_regressor(df: pd.DataFrame,
                  model_provider: Callable[[int], Model],
                  test_size: float = 0.4,
                  cross_validation: Tuple[int, Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]] = None,
                  cache_feature_matrix: bool = False,
                  test_validate_split_seed = 42,
                  hyper_parameter_space: Dict = None,
                  ) -> Fit:

    model, (df_train, df_test), trails  = _fit(df,
                                               model_provider,
                                               test_size = test_size,
                                               cross_validation = cross_validation,
                                               cache_feature_matrix = cache_feature_matrix,
                                               test_validate_split_seed = test_validate_split_seed,
                                               hyper_parameter_space=hyper_parameter_space)

    training_summary = RegressionSummary(df_train)
    test_summary = RegressionSummary(df_test)
    return Fit(model, training_summary, test_summary, trails)


def backtest_regressor(df: pd.DataFrame, model: Model) -> None:
    df = _backtest(df, model)
    return RegressionSummary(df)


def regress(df: pd.DataFrame, model: Model, tail: int = None) -> pd.DataFrame:
    return _predict(df, model, tail)



from unittest import TestCase

import numpy as np

from pandas_ml_utils.utils import unfold_parameter_space, KFoldBoostRareEvents, ReScaler


class TestUtils(TestCase):

    def test_unfold_parameter_space(self):
        """given"""
        parameter_space = {"a": range(3, 6), "b": [1, 2, 3], "c": [4, 5]}

        """when"""
        unfolded_parameter_space = unfold_parameter_space(parameter_space, {})

        """then"""
        self.assertEqual(len(unfolded_parameter_space), 3 * 3 * 2)

    def test_rare_events_boosting(self):
        """given"""
        x = np.arange(20)
        y = np.zeros(20)
        y[9] = True
        y[12] = True
        y[13] = True

        """when"""
        folds = KFoldBoostRareEvents(3, shuffle=True).split(x, y)

        """then"""
        for f, (train_idx, test_idx) in enumerate(folds):
            self.assertEqual(np.isin([9, 12, 13], train_idx).sum(), 3)


    def test_rescaling(self):
        """given"""
        array = np.array([[1, 2], [3, 4]])
        rescaler = ReScaler((array.min(), array.max()), (-1, 1))

        """when"""
        resacled = rescaler(array)
        print(resacled)

        """then"""
        self.assertEqual(resacled[0, 0], -1)
        self.assertEqual(resacled[1, 1], 1)



import logging
import sys
from typing import Tuple, Union

import numpy as np
import pandas as pd

from ..model.summary import Summary

log = logging.getLogger(__name__)


class ReinforcementSummary(Summary):

    def __init__(self, target: pd.DataFrame, agent_history: pd.DataFrame):
        self.df = target.join(agent_history).sort_index()

    def get_data_frame(self):
        return self.df

    def _repr_html_(self):
        return self._html_()._repr_html_()

    def _html_(self, width: str = '100%'):
        # only import it needed
        from vdom.helpers import div, p, img, table, tr, td, tbody, thead, th
        import matplotlib.pyplot as plt
        import base64
        import io

        return div(p("TODO"))



import logging
import traceback

import cachetools.func
import pandas as pd


def inner_join(df, join: pd.DataFrame, prefix: str = ''):
    return pd.merge(df, join.add_prefix(prefix), left_index=True, right_index=True, how='inner', sort=True)


@cachetools.func.ttl_cache(maxsize=1, ttl=10 * 60)
def fetch_yahoo(period='max', **kwargs: str):
    import yfinance as yf

    df = None
    for k, v in kwargs.items():
        px = f'{k}_'
        df_ = None

        # bloody skew index does not have any data on yahoo
        if v == '^SKEW':
            df_ = pd.read_csv('http://www.cboe.com/publish/scheduledtask/mktdata/datahouse/skewdailyprices.csv',
                              skiprows=1,
                              parse_dates=True,
                              index_col='Date') \
                .drop(['Unnamed: 2', 'Unnamed: 3'], axis=1)
        else:
            ticker = yf.Ticker(v)
            try:
                # first ty to append the most recent data
                df_ = ticker.history(period="1d", interval='1d')[-1:].combine_first(ticker.history(period=period))
            except:
                traceback.print_exc()
                logging.warning('failed to add yf.Ticker({v}).history(period="1d", interval="1d")[-1:] fallback to hist only!')
                df_ = ticker.history(period=period)

        # print some statistics
        logging.info(f'number of rows for {k} = {len(df_)}, from {df_.index[0]} to {df_.index[-1]} period={period}')

        if df is None:
            df = df_.add_prefix(px)
        else:
            df = df.inner_join(df_, prefix=px)

    # print some statistics
    logging.info(f'number of rows for joined dataframe = {len(df)}, from {df.index[0]} to {df.index[-1]}')
    return df



import pandas as pd
import numpy as np
import unittest

from keras.models import Sequential
from keras.layers import Dense, Activation, Flatten
from keras.optimizers import Adam
from rl.agents import SARSAAgent
from rl.policy import MaxBoltzmannQPolicy
from sklearn.model_selection import KFold

import pandas_ml_utils as pdu

from sklearn.neural_network import MLPClassifier, MLPRegressor

import logging
from pandas_ml_utils.analysis.correlation_analysis import _sort_correlation

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


class ComponentTest(unittest.TestCase):

    def test_fetch_yahoo(self):
        df = pd.fetch_yahoo(spy="SPY").tail()
        self.assertTrue(df["spy_Close"].sum() > 0)

    def test_correlation_matrix(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date')[["spy_Open", "spy_High", "spy_Low", "spy_Close"]]
        corr = _sort_correlation(df.corr(), recursive=True)

        self.assertListEqual(corr.columns.tolist(), ["spy_Open", "spy_High", "spy_Close", "spy_Low"])

    def test_make_train_data(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date')
        df['label'] = df["spy_Close"] > df["spy_Open"]

        x_train, x_test, y_train, y_test, index_train, index_test, min_required_data = \
            df.make_training_data(pdu.FeaturesAndLabels(features=['vix_Close'], labels=['label'], feature_lags=[0, 1, 2]))

        self.assertEqual(x_train.shape, (4022, 3, 1))
        self.assertEqual(y_train.shape, (4022, ))

        self.assertEqual(x_test.shape, (2682, 3, 1))
        self.assertEqual(y_test.shape, (2682,))

        self.assertEqual(len(x_train), len(index_train))
        self.assertEqual(len(x_test), len(index_test))

        self.assertEqual(min_required_data, 3)

    def test_fit_classifier_full(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date')
        df['label'] = df["spy_Close"] > df["spy_Open"]

        # fit
        fit = df.fit_classifier(pdu.SkitModel(MLPClassifier(activation='tanh', hidden_layer_sizes=(60, 50), alpha=0.001,
                                                            random_state=42),
                                              pdu.FeaturesAndLabels(features=['vix_Close'], labels=['label'],
                                                                    targets=("vix_Open", "spy_Volume"))),
                                test_size=0.4,
                                test_validate_split_seed=42)

        self.assertEqual(fit.model.min_required_data, 1)
        np.testing.assert_array_equal(fit.training_summary.get_confusion_matrix()['vix_Open'], np.array([[1067,  872], [1002, 1082]]))
        np.testing.assert_array_equal(fit.test_summary.get_confusion_matrix()['vix_Open'], np.array([[744, 586], [655, 698]]))

        # backtest
        fitted_model = fit.model
        backtest_classification = df.backtest_classifier(fitted_model)
        np.testing.assert_array_equal(backtest_classification.get_confusion_matrix()['vix_Open'], np.array([[1811, 1458], [1657, 1780]]))

        # classify
        fitted_model = fit.model
        classified_df = df.classify(fitted_model)
        print(classified_df.tail())

        self.assertEqual(len(classified_df[classified_df["vix_Open", "prediction", "value"] == False]), 3437)
        self.assertTrue(classified_df["vix_Open", "prediction", "value_proba"].sum() > 0)
        self.assertTrue(classified_df["vix_Open", "prediction", "value_proba"].min() > 0)
        self.assertTrue(classified_df["vix_Open", "prediction", "value_proba"].max() < 1)
        self.assertListEqual(classified_df.columns.tolist(),
                             [('vix_Open', 'prediction', 'value'), ('vix_Open', 'prediction', 'value_proba'), ('vix_Open', 'target', 'value')])

        # classify tail
        fitted_model = fit.model
        classified_df = df.classify(fitted_model, 2)
        self.assertEqual(len(classified_df), 2)

    def test_fit_classifier_simple(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date')
        df['label'] = df["spy_Close"] > df["spy_Open"]

        # fit
        fit = df.fit_classifier(pdu.SkitModel(MLPClassifier(activation='tanh', hidden_layer_sizes=(60, 50), alpha=0.001,
                                                            random_state=42),
                                              pdu.FeaturesAndLabels(features=['vix_Close'], labels=['label'])),
                                test_size=0.4,
                                test_validate_split_seed=42)

        fitted_model = fit.model
        classified_df = df.classify(fitted_model)

        self.assertListEqual(classified_df.columns.tolist(),
                             [('target', 'prediction', 'value'), ('target', 'prediction', 'value_proba'), ('target', 'target', 'value')])

    def test_fit_regressor(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date') / 50.

        # fit
        fit = df.fit_regressor(pdu.SkitModel(MLPRegressor(activation='tanh', hidden_layer_sizes=(4, 3, 2, 1, 2, 3, 4),
                                                          random_state=42),
                                             pdu.FeaturesAndLabels(
                                                 features=['vix_Open', 'vix_High', 'vix_Low', 'vix_Close'],
                                                 labels=['vix_Open', 'vix_High', 'vix_Low', 'vix_Close'])),
                               test_size=0.4,
                               test_validate_split_seed=42)

        fitted_model = fit.model

        # backtest
        backtest_regression = df.backtest_regressor(fitted_model)
        self.assertIsNotNone(backtest_regression)

        # regressed
        regressed = df.regress(fitted_model)
        print(regressed.tail())
        self.assertListEqual(regressed.columns.tolist(),
                             [('target', 'target', 'value'),
                              ('target', 'prediction', 'vix_Open'),
                              ('target', 'prediction', 'vix_High'),
                              ('target', 'prediction', 'vix_Low'),
                              ('target', 'prediction', 'vix_Close')])

        self.assertEqual(len(regressed), 6706)

    def test_cross_validation(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date')
        df['label'] = df["spy_Close"] > df["spy_Open"]

        # KFold
        cv = KFold(n_splits=10, shuffle=False)

        # fit
        fit = df.fit_classifier(pdu.SkitModel(MLPClassifier(activation='tanh', hidden_layer_sizes=(60, 50), alpha=0.001,
                                                            random_state=42, max_iter=10),
                                              pdu.FeaturesAndLabels(features=['vix_Close'],
                                                                    labels=['label'],
                                                                    targets=("vix_Open", "spy_Volume"))),
                                test_size=0.4,
                                cross_validation = (2, cv.split),
                                test_validate_split_seed=42)

        self.assertEqual(fit.model.min_required_data, 1)
        np.testing.assert_array_equal(fit.test_summary.get_confusion_matrix()["vix_Open"],
                                      np.array([[257, 169], [1142, 1115]]))

    def test_hyper_parameter(self):
        from hyperopt import hp

        df = pd.read_csv(f'{__name__}.csv', index_col='Date')
        df['label'] = df["spy_Close"] > df["spy_Open"]

        # fit with find hyper parameter
        fit = df.fit_classifier(
            pdu.SkitModel(MLPClassifier(activation='tanh', hidden_layer_sizes=(60, 50), random_state=42),
                          pdu.FeaturesAndLabels(features=['vix_Close'], labels=['label'],
                                                target_columns=["vix_Open"],
                                                loss_column="spy_Volume")),
            test_size=0.4,
            test_validate_split_seed=42,
            hyper_parameter_space={'alpha': hp.choice('alpha', [0.0001, 10]), 'early_stopping': True, 'max_iter': 50,
                                   '__max_evals': 4, '__rstate': np.random.RandomState(42)}
        )

        # test best parameter
        self.assertEqual(fit.model.skit_model.get_params()['alpha'], 0.0001)

    def test_multi_model(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date')
        df['label1'] = df["spy_Close"] > df["spy_Open"]
        df['label2'] = df["spy_Close"] > df["spy_Open"]

        # define the model
        model = pdu.SkitModel(MLPRegressor(activation='tanh', hidden_layer_sizes=(60, 50), alpha=0.001, random_state=42),
                              pdu.FeaturesAndLabels(features=['vix_Close'], labels=['vix_High', 'vix_Low'],
                                                    targets={"vix_High": ("spy_Volume", "vix_High"),
                                                             "vix_Low": ("spy_Volume", "vix_Low")}))
        # fit
        fit = df.fit_regressor(pdu.MultiModel(model),
                               test_size=0.4,
                               test_validate_split_seed=42)

        self.assertListEqual(fit.training_summary.df.columns.tolist(), fit.test_summary.df.columns.tolist())
        self.assertListEqual(fit.training_summary.df.columns.tolist(), [('vix_High', 'target', 'value'),
                                                                        ('vix_High', 'prediction', 'value'),
                                                                        ('vix_Low', 'target', 'value'),
                                                                        ('vix_Low', 'prediction', 'value'),
                                                                        ('vix_High', 'label', 'value'),
                                                                        ('vix_Low', 'label', 'value'),
                                                                        ('vix_High', 'loss', 'value'),
                                                                        ('vix_Low', 'loss', 'value')])

    @unittest.skip("we need a better model for reinforcement learning")
    def test_reinforcement_model(self):
        df = pd.read_csv(f'{__name__}.csv', index_col='Date')
        df['vix_Close'] = df['vix_Close'] / 50
        df['label'] = (df["spy_Close"] - df["spy_Open"]).shift(-1)
        df = df.drop('2019-09-13').dropna()[-100:]

        # define agent with model:
        def agent_provider(observation_space_shape, nb_actions):
            print(f"observation_space_shape: {observation_space_shape}")

            # define model
            model = Sequential()
            model.add(Flatten(input_shape=(1,) + observation_space_shape))
            model.add(Dense(16))
            model.add(Activation('sigmoid'))
            model.add(Dense(nb_actions))
            model.add(Activation('relu'))

            # define agent
            policy = MaxBoltzmannQPolicy()
            sarsa = SARSAAgent(model=model, nb_actions=nb_actions, nb_steps_warmup=0, policy=policy)
            sarsa.compile(Adam(lr=1e-4), metrics=['mae'])

            return sarsa

        # fit
        fit = df.fit_agent(pdu.OpenAiGymModel(agent_provider,
                                              pdu.FeaturesAndLabels(features=['vix_Close'],
                                                                    labels=['label'],
                                                                    feature_lags=range(5),
                                                                    label_type=float,
                                                                    target_columns=["vix_Close", "spy_Close"],
                                                                    loss_column="spy_Volume"),
                                              [
                                                  lambda y: -0.02, # do nothing
                                                  lambda y: y  # get trade reward
                                              ],
                                              [df['label'].min(), df['label'].max()],
                                              episodes=15),
                           test_size=0.4,
                           test_validate_split_seed=42)

        print(fit.training_summary.get_data_frame().tail())
        print(fit.training_summary.get_data_frame()["reward_history"].sum())
        print(fit.test_summary.get_data_frame().tail())
        print(fit.test_summary.get_data_frame()["reward_history"].sum())

        self.assertTrue(fit.training_summary.get_data_frame()["action_history"].sum() > 1)
        self.assertTrue(fit.test_summary.get_data_frame()["action_history"].sum() > 1)



import logging
import numpy as np
import pandas as pd
from typing import List, Iterable, Union, Tuple
from sklearn.ensemble import ExtraTreesClassifier, ExtraTreesRegressor
from pandas_ml_utils.analysis.correlation_analysis import plot_correlation_matrix, _sort_correlation, _plot_heatmap
log = logging.getLogger(__name__)


def feature_selection(df: pd.DataFrame,
                      label_column: str = None,
                      ignore: Union[List[str], str] = [],
                      top_features: int = 5,
                      correlation_threshold: float = 0.5,
                      minimum_features: int = 1,
                      lags: Iterable[int] = range(100),
                      show_plots: bool = True,
                      figsize: Tuple[int] = (12, 10)):
    """
    The *feature_selection* functionality helps you to analyze your features, filter out highly
    correlated once and focus on the most important features. This function also applies an
    auto regression and embeds and ACF plot.

    :param df: the DataFrame which you apply the function on
    :param label_column: column name of your dependent variable
    :param ignore: columns you want to ignore
    :param top_features: number of most important features you want to select
    :param correlation_threshold: threshold at which correlated features drop out
    :param minimum_features: number of features you want to keep even if they are highly correlated
    :param lags: iterable of lags you want to analyze as an AR process
    :param show_plots: whether to show plots or not
    :param figsize: size of the polots
    :return: None
    """
    df = df.drop(ignore, axis=1).dropna()
    N = len(df)
    correlation_mx = _sort_correlation(df.corr())

    if show_plots:
        plot_correlation_matrix(df)

    print(correlation_mx[:1])
    features = correlation_mx

    # find the features most correlated to the label (if provided)
    if label_column is not None:
        # drop label column
        features = features.drop(label_column, axis=0)
        features = features.drop(label_column, axis=1)

        estimators, importances, indices, imporant_feature_names = \
            __feature_importance(df[features.columns].values, df[label_column].values, features.columns)

        if show_plots:
            __plot_feature_importance(len(features.columns), estimators, importances, indices, imporant_feature_names, figsize)

        print(f"Feature ranking:\n{imporant_feature_names.tolist()}")

        features = df[imporant_feature_names[:min(top_features, len(imporant_feature_names))]].corr()

        print(f"\nTOP {top_features} features")
        print(features[:1])

    # then eliminate features with high correlation to each other
    features = _sort_correlation(features)
    while len(features) > minimum_features and __max_correlation(features) > correlation_threshold:
        index = np.unravel_index(__argmax_correlation(features), features.values.shape)
        features = features.drop(features.index[index[1]], axis=0)
        features = features.drop(features.columns[index[1]], axis=1)

    print(f"\nfiltered features with correlation < {correlation_threshold}")
    print(correlation_mx[features.columns][:1])

    if show_plots:
        _plot_heatmap(features, figsize)

    # make AR analysis of remaining features
    coefficients = []
    for feature in features.columns:
        if show_plots:
            __plot_acf(df[feature], lags, (figsize[0], int(figsize[1] / 3)))

        dff = df[[feature]]
        for lag in lags:
            dff = dff.join(df[feature].shift(lag), rsuffix=f'_{lag}')

        dff = dff.corr().dropna()
        acorr = dff.iloc[0]
        coefficients.append(acorr.values)
        print(acorr.sort_values(ascending=False)[:10])

    # find clusters of auto correlation coefficients over all features
    cl = len(coefficients)
    best_lags = np.array(coefficients).sum(axis=0)
    best_lags_i = (-best_lags).argsort()
    best_lags = [(i, f'{-best_lags[i] / cl:.2f}') for i in best_lags_i if abs(best_lags[i]) > (1.96 / np.sqrt(N - i) / cl)]
    print(f"best lags are\n{best_lags[1:]}")


def __feature_importance(x, y, names, n_estimators=250):
    continuous = 'float' in (str(y.dtype))
    print(f"label is continuous: {continuous}")

    forest = ExtraTreesRegressor(n_estimators=n_estimators, random_state=0) if continuous \
             else ExtraTreesClassifier(n_estimators=n_estimators, random_state=0)

    forest.fit(x, y)

    importances = forest.feature_importances_
    indices = np.argsort(importances)[::-1]

    # return ranking
    return forest.estimators_, importances, indices, names[indices]


def __plot_feature_importance(nr_of_features, estimators, importances, indices, names, figsize):
    std = np.std([tree.feature_importances_ for tree in estimators], axis=0)

    # Plot the feature importances of the forest
    try:
        from matplotlib import pyplot as plt
        plt.figure(figsize=figsize)
        plt.title("Feature importances")
        plt.bar(range(nr_of_features), importances[indices], color="r", yerr=std[indices], align="center")
        plt.xticks(range(nr_of_features), names, rotation='vertical')
        plt.xlim([-1, nr_of_features])
        plt.show()
    except:
        return None


def __plot_acf(series, lags, figsize):
    try:
        from matplotlib import pyplot as plt
        from statsmodels.graphics.tsaplots import plot_acf

        fig, ax = plt.subplots(figsize=figsize)
        fig.suptitle(series.name, fontsize=14)

        plot_acf(series, lags=lags, ax=ax)
        plt.show()
    except:
        return None


def __max_correlation(df):
    return (df.abs().values - np.identity(len(df.columns))).max()


def __argmax_correlation(df):
    return (df.abs().values - np.identity(len(df.columns))).argmax()


import logging
import os
import sys
from functools import lru_cache
from typing import Tuple, Union, Dict

import numpy as np
import pandas as pd

from ..constants import *
from ..model.summary import Summary
from sklearn.metrics import f1_score

log = logging.getLogger(__name__)


class ClassificationSummary(Summary):
    def __init__(self, df: pd.DataFrame, probability_cutoff: float = 0.5):
        super().__init__()
        self.df = df
        self.probability_cutoff = probability_cutoff
        self.confusions = {target: ClassificationSummary._calculate_confusions(df[target]) for target in {*df.columns.get_level_values(0)}}

    @lru_cache(maxsize=None)
    def get_confusion_matrix(self):
        return {target: np.array([[len(c) for c in r] for r in cm]) for target, cm in self.confusions.items()}

    @lru_cache(maxsize=None)
    def get_confusion_loss(self):
        return {target: np.array([[c[LOSS_COLUMN_NAME, "value"].sum() for c in r] for r in cm]) for target, cm in self.confusions.items()}

    @lru_cache(maxsize=None)
    def get_metrics(self):
        return {target: {"FP Ratio": fp_ratio,
                         "FN Ratio": fn_ratio,
                         "F1 Score": f1_score(self.df[target, LABEL_COLUMN_NAME, "value"],
                                              self.df[target, PREDICTION_COLUMN_NAME, "value"])}
                for target, (fp_ratio, fn_ratio) in self.get_ratios().items()}

    @lru_cache(maxsize=None)
    def get_ratios(self):
        return {target: (cm[0,1] / cm[0,0], cm[1,0] / cm[0,0]) for target, cm in self.get_confusion_matrix().items()}

    @lru_cache(maxsize=None)
    def plot_classification(self, figsize=(16, 9)) -> Dict:
        import seaborn as sns
        import matplotlib.pyplot as plt
        from matplotlib import gridspec
        from pandas.plotting import register_matplotlib_converters

        probability_cutoff = self.probability_cutoff
        plots = {}

        for target in self.df.columns.levels[0]:
            df = self.df[target]

            # get rid of deprecation warning
            register_matplotlib_converters()

            # define grid
            fig = plt.figure(figsize=figsize)
            gs = gridspec.GridSpec(2, 1, height_ratios=[1, 3])
            ax0 = plt.subplot(gs[0])
            ax1 = plt.subplot(gs[1])

            # plot probability
            bar = sns.lineplot(x=range(len(df)), y=df[PREDICTION_COLUMN_NAME, f"value{PROBABILITY_POSTFIX}"], ax=ax0)
            ax0.hlines(probability_cutoff, 0, len(df), color=sns.xkcd_rgb['silver'])

            # plot loss
            color = pd.Series(0, index=df.index)
            color.loc[df[(PREDICTION_COLUMN_NAME, "value")] & df[(LABEL_COLUMN_NAME, "value")]] = 1
            color.loc[(df[(PREDICTION_COLUMN_NAME, "value")] == False) & df[(LABEL_COLUMN_NAME, "value")]] = 2

            colors = {0: sns.xkcd_rgb['white'], 1: sns.xkcd_rgb['pale green'], 2: sns.xkcd_rgb['cerise']}
            palette = [colors[color_index] for color_index in np.sort(color.unique())]

            scatt = sns.scatterplot(ax=ax1,
                                    x=range(len(df)),
                                    y=df[(LOSS_COLUMN_NAME, "value")].clip(upper=0),
                                    size=df[(LOSS_COLUMN_NAME, "value")] * -1,
                                    hue=color,
                                    palette=palette)

            plt.close()
            plots[target] = fig

        return plots

    @staticmethod
    def _calculate_confusions(df):
        tp = df[df[PREDICTION_COLUMN_NAME, "value"] & df[LABEL_COLUMN_NAME, "value"]]
        fp = df[df[PREDICTION_COLUMN_NAME, "value"] & (df[LABEL_COLUMN_NAME, "value"] == False)]
        tn = df[(df[PREDICTION_COLUMN_NAME, "value"] == False) & (df[LABEL_COLUMN_NAME, "value"] == False)]
        fn = df[(df[PREDICTION_COLUMN_NAME, "value"] == False) & (df[LABEL_COLUMN_NAME, "value"])]

        return [[tp, fp],
                [fn, tn]]

    @lru_cache(maxsize=None)
    def _repr_html_(self):
        from mako.template import Template
        from mako.lookup import TemplateLookup

        template = Template(filename=f"{os.path.abspath(__file__)}.html", lookup=TemplateLookup(directories=['/']))
        return template.render(classification_summary=self)


import logging
from typing import Callable, Tuple, Dict

import numpy as np
import pandas as pd

from ..model.fit import Fit
from ..model.fitter import _fit, _predict
from ..model.models import Model
from ..train_test_data import make_training_data
from .summary import ReinforcementSummary

log = logging.getLogger(__name__)


def fit_agent(df: pd.DataFrame,
              model_provider: Callable[[int], Model],
              test_size: float = 0.4,
              cross_validation: Tuple[int, Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]] = None,
              cache_feature_matrix: bool = False,
              test_validate_split_seed = 42,
              hyper_parameter_space: Dict = None,
              ) -> Fit:

    model, train, test, index, prediction, trails = _fit(df,
                                                         model_provider,
                                                         test_size = test_size,
                                                         cross_validation = cross_validation,
                                                         cache_feature_matrix = cache_feature_matrix,
                                                         test_validate_split_seed = test_validate_split_seed,
                                                         hyper_parameter_space=hyper_parameter_space)

    target_columns = [target for target, _ in model.features_and_labels.get_goals().items()]
    train_targets = df[target_columns].loc[index[0]]
    test_targets = df[target_columns].loc[index[1]]

    training_classification = ReinforcementSummary(train_targets, model.history[0])
    test_classification = ReinforcementSummary(test_targets, model.history[1])
    return Fit(model, training_classification, test_classification, trails)


def backtest_agent(df: pd.DataFrame, model: Model) -> ReinforcementSummary:
    features_and_labels = model.features_and_labels

    # make training and test data with no 0 test data fraction
    x, _, y, _, index, _, _ = make_training_data(df, features_and_labels, 0, int)

    target_columns = [target for target, _ in model.features_and_labels.get_goals().items()]
    targets = df[target_columns]
    back_test_history = model.back_test(index, x, y)

    return ReinforcementSummary(targets, back_test_history)


def agent_take_action(df: pd.DataFrame, model: Model, tail: int = None) -> pd.DataFrame:
    dff = _predict(df, model, tail)
    return dff



import logging
from typing import Callable, Tuple, Dict

import numpy as np
import pandas as pd

from ..classification.summary import ClassificationSummary
from ..model.fit import Fit
from ..model.fitter import _fit, _backtest, _predict
from ..model.models import Model
from ..constants import *

log = logging.getLogger(__name__)


def fit_classifier(df: pd.DataFrame,
                   model_provider: Callable[[int], Model],
                   test_size: float = 0.4,
                   cross_validation: Tuple[int, Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]] = None,
                   cache_feature_matrix: bool = False,
                   test_validate_split_seed = 42,
                   hyper_parameter_space: Dict = None,
                   ) -> Fit:
    """

    :param df: the DataFrame you apply this function to
    :param model_provider: a callable which provides a new :class:`.Model` instance i.e. for each hyper parameter if
                           hyper parameter tuning is enforced. Usually all the Model subclasses implement __call__
                           thus they are a provider of itself
    :param test_size: the fraction [0, 1] of samples which are used for a test set
    :param cross_validation: tuple of number of epochs for each fold provider and a cross validation provider
    :param cache_feature_matrix: whether to cache an expensive generation of feature matrices between fit calls
    :param test_validate_split_seed: seed if train, test split needs to be reproduceable
    :param hyper_parameter_space: space of hyper parameters passed as kwargs to your model provider
    :return: returns a :class:`.Fit` object
    """
    # maybe later we can check if only the cut off changed and then skip the fitting step
    model, (df_train, df_test), trails = _fit(df,
                                              model_provider,
                                              test_size = test_size,
                                              cross_validation = cross_validation,
                                              cache_feature_matrix = cache_feature_matrix,
                                              test_validate_split_seed = test_validate_split_seed,
                                              hyper_parameter_space = hyper_parameter_space)

    # assemble the result objects
    cutoff = model[("probability_cutoff", 0.5)]

    # convert probabilities into classes
    df_train = _convert_probabilities(df_train, cutoff)
    df_test = _convert_probabilities(df_test, cutoff)

    training_classification = ClassificationSummary(df_train, cutoff)
    test_classification = ClassificationSummary(df_test, cutoff)
    return Fit(model, training_classification, test_classification, trails)


def backtest_classifier(df: pd.DataFrame, model: Model) -> ClassificationSummary:
    df = _backtest(df, model)
    df_sumary = _convert_probabilities(df)
    return ClassificationSummary(df_sumary, model[("probability_cutoff", 0.5)])


def classify(df: pd.DataFrame, model: Model, tail: int = None) -> pd.DataFrame:
    dff = _predict(df, model, tail)
    return _convert_probabilities(dff, model[("probability_cutoff", 0.5)])


def _convert_probabilities(df: pd.DataFrame, cut_off: float = 0.5) -> pd.DataFrame:

    # return result
    for column in df.columns:
        if column[1] == PREDICTION_COLUMN_NAME:
            probability_column = (column[0], column[1], f"{column[2]}{PROBABILITY_POSTFIX}")
            df[probability_column] = df[column]
            df[column] = df[probability_column] > cut_off

    return df.sort_index(axis=1)



import numpy as np
import pandas as pd
import logging

from unittest import TestCase
from pandas_ml_utils.classification.classifier import _convert_probabilities

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

target = [0.0, 0.0, 0.0, 0.0, 0.0]
loss = np.array([1.0, 1.0, 1.0, 1.0, 1.0]) * -1

df = pd.DataFrame({
    ("no fit", "target", "value"):          target,
    ("no fit", "prediction", "value"):      [1.0, 0.0, 1.0, 1.0, 0.0],
    ("no fit", "label", "value"):           [False, True, False, False, True],
    ("no fit", "loss", "value"):            loss
})


class TestClassifier(TestCase):

    def test__convert_probabilities(self):
        """given"""
        cut_off = 0.5

        """when"""
        augmented = _convert_probabilities(df, cut_off)

        """then"""
        self.assertTrue(("no fit", "prediction", "value_proba") in augmented.columns)
        self.assertEqual(augmented[("no fit", "prediction", "value_proba")].dtype, float)
        self.assertEqual(augmented[("no fit", "prediction", "value")].dtype, bool)



import logging
import re
from typing import Union, Callable, List

import numpy as np
import pandas as pd

log = logging.getLogger(__name__)


def add_apply(df, **kwargs: Callable[[pd.DataFrame], Union[pd.Series, pd.DataFrame]]):
    df2 = pd.DataFrame()
    for k, v in kwargs.items():
        df2[k] = v(df)

    return df.join(df2)


def shift_inplace(df, **kwargs: int):
    for k, v in kwargs.items():
        df[k] = df[k].shift(v)

    return df


def drop_re(df, *args: str):
    drop_list = []

    for regex in args:
        drop_list.extend(list(filter(re.compile(regex).match, df.columns)))

    return df.drop(drop_list, axis=1)


def drop_zero_or_nan(df, columns: List[str] = None, epsilon=1e-10):
    df = df.copy()

    # replace all zeros by nan
    for col in columns if columns is not None else df.columns:
        df[df[col].abs() < epsilon] = np.NaN

    # then drop all nan
    return df.dropna()


def extend_forecast(df, periods: int):
    df_ext = pd.DataFrame(index=pd.date_range(df.index[-1], periods=periods+1, closed='right'))
    return pd.concat([df, df_ext], axis=0, sort=True)



import os
from functools import lru_cache

import numpy as np
import pandas as pd
import logging

from time import perf_counter as pc
from sortedcontainers import SortedDict
from typing import Type, Callable, Tuple

from pandas_ml_utils.wrappers.hashable_dataframe import HashableDataFrame
from pandas_ml_utils.utils import log_with_time, ReScaler

log = logging.getLogger(__name__)


def make_backtest_data(df: pd.DataFrame,
                       features_and_labels: 'FeaturesAndLabels',
                       label_type: Type = int):
    return make_training_data(df, features_and_labels, 0, label_type)


def make_training_data(df: pd.DataFrame,
                       features_and_labels: 'FeaturesAndLabels',
                       test_size: float = 0.4,
                       label_type: Type = int,
                       seed: int = 42,
                       cache: bool = False) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, list, list, int]:
    # only import if this method is needed
    from sklearn.model_selection import train_test_split

    # use only feature and label columns
    start_pc = log_with_time(lambda: log.debug("make training / test data split ..."))
    df = df[set(features_and_labels.features + features_and_labels.labels)]

    # create features and re-assign data frame with all nan rows dropped
    df_new, x = _make_features_with_cache(HashableDataFrame(df), features_and_labels) if cache else \
                _make_features(df, features_and_labels)

    # calculate the minimum required data
    min_required_data = len(df) - len(df_new) + 1

    # assign labels
    y = df_new[features_and_labels.labels].values

    # split training and test data
    start_split_pc = log_with_time(lambda: log.debug("  splitting ..."))
    x_train, x_test, y_train, y_test, index_train, index_test = \
        train_test_split(x, y, df_new.index, test_size=test_size, random_state=seed) if test_size > 0 \
            else (x, None, y, None, df_new.index, None)
    log.info(f"  splitting ... done in {pc() - start_split_pc: .2f} sec!")

    # ravel one dimensional labels
    if len(features_and_labels.labels) == 1:
        y_train = y_train.ravel().astype(label_type)
        y_test = y_test.ravel().astype(label_type) if y_test is not None else None

    log.info(f"make training / test data split ... done in {pc() - start_pc: .2f} sec!")

    # return the split
    #log.debug(f"${len(x_train)}, {len(x_test)}, {len(y_train)}, {len(y_test)}, {len(index_train)}, {len(index_test)}, {min_required_data}")
    return x_train, x_test, y_train, y_test, index_train, index_test, min_required_data


def make_forecast_data(df: pd.DataFrame, features_and_labels: 'FeaturesAndLabels'):
    return _make_features(df[features_and_labels.features], features_and_labels)


@lru_cache(maxsize=int(os.getenv('CACHE_FEATUES_AND_LABELS', '1')))
def _make_features_with_cache(df: HashableDataFrame, features_and_labels: 'FeaturesAndLabels'):
    log.info(f"no cache entry available for {hash(df), hash(features_and_labels)}")
    return _make_features(df, features_and_labels)


def _make_features(df: pd.DataFrame, features_and_labels: 'FeaturesAndLabels'):
    start_pc = log_with_time(lambda: log.debug(" make features ..."))
    feature_lags = features_and_labels.feature_lags
    feature_rescaling = features_and_labels.feature_rescaling
    features = features_and_labels.features
    lag_smoothing = features_and_labels.lag_smoothing

    # drop nan's and copy frame
    df = df.dropna().copy()

    # generate feature matrix
    if feature_lags is not None:
        # return RNN shaped 3D arrays
        for feature in features:
            feature_series = df[feature]
            smoothers = None

            # smooth out feature if requested
            if lag_smoothing is not None:
                smoothers = SortedDict({lag: smoother(feature_series.to_frame())
                                        for lag, smoother in lag_smoothing.items()})

            for lag in feature_lags:
                # if smoothed values are applicable use smoothed values
                if smoothers is not None and len(smoothers) > 0 and smoothers.peekitem(0)[0] <= lag:
                    feature_series = smoothers.popitem(0)[1]

                # assign the lagged (eventually smoothed) feature to the features frame
                df[f'{feature}_{lag}'] = feature_series.shift(lag)

        # drop all rows which got nan now
        df = df.dropna()

        # RNN shape need to be [row, time_step, feature]
        columns = features_and_labels.get_feature_names()
        x = np.array([df[cols].values for cols in columns], ndmin=3).swapaxes(0, 1)
    else:
        # return simple 2D arrays
        x = df[features].values

    if feature_rescaling is not None:
        for tuple, target_range in feature_rescaling.items():
            for i in range(len(x)):
                cols = [features.index(j) for j in tuple]
                if len(x.shape) == 3:
                    data = x[i, :, cols]
                    x[i,:,cols] = ReScaler((data.min(), data.max()), target_range)(data)
                elif len(x.shape) == 2:
                    data = x[i,cols]
                    x[i,cols] = ReScaler((data.min(), data.max()), target_range)(data)
                else:
                    ValueError("unknown array dimensions")

    log.info(f" make features ... done in {pc() - start_pc: .2f} sec!")
    return df, x


def reshape_rnn_as_ar(arr3d):
    if len(arr3d.shape) < 3:
        print("Data was not in RNN shape")
        return arr3d
    else:
        return arr3d.reshape(arr3d.shape[0], arr3d.shape[1] * arr3d.shape[2])




import pandas as pd
import numpy as np
import gym
from gym import spaces
from typing import Tuple, Callable, List

from ..model.features_and_Labels import FeaturesAndLabels

INIT_ACTION = -1


class RowWiseGym(gym.Env):

    def __init__(self,
                 environment: Tuple[np.ndarray, np.ndarray, np.ndarray],
                 features_and_labels: FeaturesAndLabels,
                 action_reward_functions: List[Callable[[np.ndarray], float]],
                 observation_range: Tuple[int, int],
                 reward_range: Tuple[int, int]):
        super().__init__()
        self.environment = environment
        self.reward_range = reward_range
        self.action_reward_functions = action_reward_functions

        # start at the beginning of the frame
        self.state = 0

        # define spaces
        self.action_space = spaces.Discrete(len(action_reward_functions))
        self.observation_space = spaces.Box(low=observation_range[0], high=observation_range[1],
                                            shape=features_and_labels.shape()[0], dtype=np.float16)

        # define history
        self.reward_history = []
        self.action_history = []

    metadata = {'render.modes': ['human']}

    def reset(self):
        # Reset the state of the environment to an initial state and reset history
        self.reward_history = []
        self.action_history = []
        return self.step(INIT_ACTION)[0]

    def step(self, action):
        # Execute one time step within the environment
        if action is not INIT_ACTION:
            reward = self.action_reward_functions[action](self.environment[2][self.state])
            self.reward_history.append(reward)
            self.action_history.append(action)
            self.state += 1
        else:
            reward = 0
            self.state = 0

        done = self.state >= len(self.environment[1])
        obs = self.environment[1][self.state if not done else 0]

        return obs, reward, done, {}

    def render(self, mode='human', close=False):
        print(f"reward: {sum(self.reward_history)}")

    def get_history(self):
        return pd.DataFrame({"reward_history": self.reward_history,
                             "action_history": self.action_history},
                            index=self.environment[0]).sort_index()


import pandas as pd
from unittest import TestCase

from pandas_ml_utils.model.models import *
from pandas_ml_utils.model.fitter import _fit, _backtest, _predict

from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.svm import LinearSVC


df = pd.DataFrame({"a": np.array([0.1, 0.01]), "b": np.array([True, False]), "c": np.array([False, True])})

class TestFitter(TestCase):

    def test__fit(self):
        """given"""
        features_and_labels = FeaturesAndLabels(["a"], ["b"])
        providers = [
            SkitModel(MLPClassifier(activation='tanh', hidden_layer_sizes=(1, 1), alpha=0.001, random_state=42),
                      features_and_labels, foo='bar'),
            SkitModel(LogisticRegression(), features_and_labels),
            SkitModel(LinearSVC(), features_and_labels),
            SkitModel(RandomForestClassifier(), features_and_labels)]

        """when"""
        fitts = [_fit(df, p, 0)[1][0] for p in providers]
        fits_df_columns = [f.columns.tolist() for f in fitts]

        """then"""
        expected_columns = [('target', 'target', 'value'), ('target', 'prediction', 'value'), ('target', 'label', 'value'), ('target', 'loss', 'value')]
        self.assertListEqual(fits_df_columns[0], expected_columns)
        self.assertListEqual(fits_df_columns[1], expected_columns)
        self.assertListEqual(fits_df_columns[2], expected_columns)
        self.assertListEqual(fits_df_columns[3], expected_columns)
        np.testing.assert_array_equal(fitts[0]["target", "label", "value"].values, df["b"].values)
        np.testing.assert_array_equal(fitts[1]["target", "label", "value"].values, df["b"].values)
        np.testing.assert_array_equal(fitts[2]["target", "label", "value"].values, df["b"].values)
        np.testing.assert_array_equal(fitts[3]["target", "label", "value"].values, df["b"].values)

    def test__backtest(self):
        """given"""
        fls = [FeaturesAndLabels(["a"], ["b"]),
               FeaturesAndLabels(["a"], ["b"], targets="b"),
               FeaturesAndLabels(["a"], ["b", "c"], targets={"b": (-1, ["b", "c"])}),
               FeaturesAndLabels(["a"], ["b", "c"], targets={"b": (-1, ["b", "c"]), "a": (-2, ["b", "c"])})]

        providers = [SkitModel(MLPRegressor(activation='tanh', hidden_layer_sizes=(1, 1), alpha=0.001, random_state=42),
                               features_and_labels=fl) for fl in fls]

        """when"""
        fitted_models = [_fit(df, p, 0)[0] for p in providers]
        backtests = [_backtest(df, fm) for fm in fitted_models]
        backtest_columns = [b.columns.tolist() for b in backtests]

        """then"""
        # print(backtest_columns[3])
        self.assertEqual(backtest_columns[0], [('target', 'target', 'value'), ('target', 'prediction', 'value'), ('target', 'label', 'value'), ('target', 'loss', 'value')])
        self.assertEqual(backtest_columns[1], [('b', 'target', 'value'), ('b', 'prediction', 'value'), ('b', 'label', 'value'), ('b', 'loss', 'value')])
        self.assertEqual(backtest_columns[2], [('b', 'target', 'value'), ('b', 'prediction', 'b'), ('b', 'prediction', 'c'), ('b', 'label', 'b'), ('b', 'label', 'c'), ('b', 'loss', 'value')])
        self.assertEqual(backtest_columns[3], [('b', 'target', 'value'), ('b', 'prediction', 'b'), ('b', 'prediction', 'c'), ('a', 'target', 'value'), ('a', 'prediction', 'b'), ('a', 'prediction', 'c'), ('b', 'label', 'b'), ('b', 'label', 'c'), ('a', 'label', 'b'), ('a', 'label', 'c'), ('b', 'loss', 'value'), ('a', 'loss', 'value')])
        np.testing.assert_array_equal(backtests[3]["b", "label", "b"].values, df["b"].values)
        np.testing.assert_array_equal(backtests[3]["a", "label", "b"].values, df["b"].values)
        np.testing.assert_array_equal(backtests[3]["b", "label", "c"].values, df["c"].values)
        np.testing.assert_array_equal(backtests[3]["a", "label", "c"].values, df["c"].values)
        np.testing.assert_array_equal(backtests[3]["a", "loss", "value"].values, -2)
        np.testing.assert_array_equal(backtests[3]["b", "loss", "value"].values, -1)

    def test__predict(self):
        """given"""
        fls = [FeaturesAndLabels(["a"], ["b"]),
               FeaturesAndLabels(["a"], ["b"], targets="b"),
               FeaturesAndLabels(["a"], ["c", "b"], targets={"b": (-1, ["b", "c"])}),
               FeaturesAndLabels(["a"], ["c", "b"], targets={"b": (-1, ["b", "c"]), "a": (-1, ["b", "c"])})]

        providers = [SkitModel(MLPRegressor(activation='tanh', hidden_layer_sizes=(1, 1), alpha=0.001, random_state=42),
                               features_and_labels=fl) for fl in fls]


        """when"""
        fitted_models = [_fit(df, p, 0)[0] for p in providers]

        """then"""
        predictions = [_predict(df, fm) for fm in fitted_models]
        print(predictions[-1].columns.tolist())
        self.assertEqual(predictions[0].columns.tolist(), [('target', 'target', 'value'), ('target', 'prediction', 'value')])
        self.assertEqual(predictions[1].columns.tolist(), [('b', 'target', 'value'), ('b', 'prediction', 'value')])
        self.assertEqual(predictions[2].columns.tolist(), [('b', 'target', 'value'), ('b', 'prediction', 'b'), ('b', 'prediction', 'c')])
        self.assertEqual(predictions[3].columns.tolist(), [('b', 'target', 'value'), ('b', 'prediction', 'b'), ('b', 'prediction', 'c') , ('a', 'target', 'value'), ('a', 'prediction', 'b'), ('a', 'prediction', 'c')])

    def test__predict_with_lags(self):
        """given"""
        df = pd.DataFrame({"a": [0.5592344, 0.60739384, 0.19994533, 0.56642537, 0.50965677,
                                 0.168989, 0.94080671, 0.76651769, 0.8403563, 0.4003567,
                                 0.24295908, 0.50706317, 0.66612371, 0.4020924, 0.21776017,
                                 0.32559497, 0.12721287, 0.13904584, 0.65887554, 0.08830925],
                           "b": range(20)})

        fl = FeaturesAndLabels(["a"], ["b"], feature_lags=[0,1,2])
        provider = SkitModel(MLPRegressor(activation='tanh', hidden_layer_sizes=(1, 1), alpha=0.001, random_state=42),
                             features_and_labels=fl)

        """when"""
        model, summaries, trails = _fit(df, provider, 0)
        backtest = _backtest(df, model)

        """then"""
        predictions = _predict(df, model)
        self.assertListEqual(predictions.columns.tolist(), [('target', 'target', 'value'), ('target', 'prediction', 'value')])
        self.assertEqual(model.min_required_data, 3)

        self.assertListEqual(backtest.columns.tolist(), [('target', 'target', 'value'), ('target', 'prediction', 'value'), ('target', 'label', 'value'), ('target', 'loss', 'value')])
        self.assertEqual(backtest["target", "loss", "value"].sum(), -18)
        self.assertEqual(backtest["target", "label", "value"].sum(), 189)




from __future__ import annotations

import logging
from time import perf_counter
from typing import Callable, Tuple, Dict, TYPE_CHECKING
from sklearn.utils.testing import ignore_warnings
from sklearn.exceptions import ConvergenceWarning

import numpy as np
import pandas as pd

from ..train_test_data import make_training_data, make_forecast_data
from ..utils import log_with_time
from ..model.models import Model
from ..constants import *

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from hyperopt import Trials


def _fit(df: pd.DataFrame,
         model_provider: Callable[[int], Model],
         test_size: float = 0.4,
         cross_validation: Tuple[int, Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]] = None,
         cache_feature_matrix: bool = False,
         test_validate_split_seed = 42,
         hyper_parameter_space: Dict = None,
         ) -> Tuple[Model, Tuple[pd.DataFrame, pd.DataFrame], Trials]:
    # get a new model
    trails = None
    model = model_provider()
    features_and_labels = model.features_and_labels
    goals = features_and_labels.get_goals()

    # make training and test data sets
    x_train, x_test, y_train, y_test, index_train, index_test, min_required_data = \
        make_training_data(df,
                           features_and_labels,
                           test_size,
                           label_type=features_and_labels.label_type,
                           seed=test_validate_split_seed,
                           cache=cache_feature_matrix)

    log.info(f"create model (min required data = {min_required_data}")
    model.min_required_data = min_required_data

    # eventually perform a hyper parameter optimization first
    start_performance_count = log_with_time(lambda: log.info("fit model"))
    if hyper_parameter_space is not None:
        # next isolate hyperopt parameters and constants only used for hyper parameter tuning like early stopping
        constants = {}
        hyperopt_params = {}
        for k, v in list(hyper_parameter_space.items()):
            if k.startswith("__"):
                hyperopt_params[k[2:]] = hyper_parameter_space.pop(k)
            elif type(v) in [int, float, bool]:
                constants[k] = hyper_parameter_space.pop(k)

        # optimize hyper parameters
        model, trails = __hyper_opt(hyper_parameter_space,
                                    hyperopt_params,
                                    constants,
                                    model_provider,
                                    cross_validation,
                                    x_train, y_train, index_train,
                                    x_test, y_test, index_test)

    # finally train the model with eventually tuned hyper parameters
    __train_loop(model, cross_validation, x_train, y_train, index_train, x_test, y_test, index_test)

    log.info(f"fitting model done in {perf_counter() - start_performance_count: .2f} sec!")
    header = __stack_header_prediction(goals) + __stack_header_label(goals) + __stack_header_loss(goals)

    df_train = df.loc[index_train]
    df_prediction_train = __predict(df_train, model, x_train) \
        .join(__truth(df_train, model)) \
        .join(__loss(df_train, model))
    df_prediction_train.columns = pd.MultiIndex.from_tuples(header)

    df_prediction_test = None
    if x_test is not None:
        df_test = df.loc[index_test]
        df_prediction_test = __predict(df_test, model, x_test) \
            .join(__truth(df_test, model)) \
            .join(__loss(df_test, model))
        df_prediction_test.columns = pd.MultiIndex.from_tuples(header)

    return model, (df_prediction_train, df_prediction_test), trails


def __train_loop(model, cross_validation, x_train, y_train, index_train,  x_test, y_test, index_test):
    if cross_validation is not None and isinstance(cross_validation, Tuple) and callable(cross_validation[1]):
        losses = []
        for fold_epoch in range(cross_validation[0]):
            # cross validation, make sure we re-shuffle every fold_epoch
            for f, (train_idx, test_idx) in enumerate(cross_validation[1](x_train, y_train)):
                log.info(f'fit fold {f}')
                loss = model.fit(x_train[train_idx], y_train[train_idx], x_train[test_idx], y_train[test_idx],
                                 index_train[train_idx], index_train[test_idx])

                losses.append(loss)

        return np.array(losses).mean()
    else:
        # fit without cross validation
        return model.fit(x_train, y_train, x_test, y_test, index_train, index_test)


@ignore_warnings(category=ConvergenceWarning)
def __hyper_opt(hyper_parameter_space,
                hyperopt_params,
                constants,
                model_provider,
                cross_validation,
                x_train, y_train, index_train,
                x_test, y_test, index_test):
    from hyperopt import fmin, tpe, Trials

    keys = list(hyper_parameter_space.keys())

    def f(args):
        sampled_parameters = {k: args[i] for i, k in enumerate(keys)}
        model = model_provider(**sampled_parameters, **constants)
        loss = __train_loop(model, cross_validation, x_train, y_train, index_train, x_test, y_test, index_test)
        if loss is None:
            raise ValueError("Can not hyper tune if model loss is None")

        return {'status': 'ok', 'loss': loss, 'parameter': sampled_parameters}

    trails = Trials()
    fmin(f, list(hyper_parameter_space.values()), algo=tpe.suggest, trials=trails, show_progressbar=False, **hyperopt_params)

    # find the best parameters amd make sure to NOT pass the constants as they are only used for hyperopt
    best_parameters = trails.best_trial['result']['parameter']
    best_model = model_provider(**best_parameters)

    print(f'best parameters: {repr(best_parameters)}')
    return best_model, trails


def _backtest(df: pd.DataFrame, model: Model) -> pd.DataFrame:
    features_and_labels = model.features_and_labels
    goals = features_and_labels.get_goals()

    # make training and test data with no 0 test data fraction
    x, _, y, _, index, _, _ = make_training_data(df, features_and_labels, 0, int)

    # predict probabilities
    df_source = df.loc[index]
    df_backtest = __predict(df_source, model, x) \
        .join(__truth(df_source, model)) \
        .join(__loss(df_source, model))

    header = __stack_header_prediction(goals) + __stack_header_label(goals) + __stack_header_loss(goals)
    df_backtest.columns = pd.MultiIndex.from_tuples(header)

    return df_backtest


def _predict(df: pd.DataFrame, model: Model, tail: int = None) -> pd.DataFrame:
    features_and_labels = model.features_and_labels
    goals = features_and_labels.get_goals()

    if tail is not None:
        if tail <= 0:
            raise ValueError("tail must be > 0 or None")
        elif model.min_required_data is not None:
            # just use the tail for feature engineering
            df = df[-(tail + (model.min_required_data - 1)):]
        else:
            log.warning("could not determine the minimum required data from the model")

    # predict and return data frame
    dff, x = make_forecast_data(df, features_and_labels)
    df_prediction = __predict(df.loc[dff.index], model, x)

    header = __stack_header_prediction(goals)
    df_prediction.columns = pd.MultiIndex.from_tuples(header)

    return df_prediction


def __predict(df, model, x):
    # first save target columns and loss column
    goals = model.features_and_labels.get_goals()
    predictions = model.predict(x)
    df_pred = pd.DataFrame({}, index=df.index)

    for target, (_, labels) in goals.items():
        prediction = predictions[target]
        postfix = "" if target is None else f'_{target}'

        if target is not None:
            df_pred[f'{TARGET_COLUMN_NAME}_{target}'] = df[target]
        else:
            df_pred[f"{TARGET_COLUMN_NAME}"] = ""

        if len(labels) > 1:
            for i, label in enumerate(labels):
                df_pred[f"{PREDICTION_COLUMN_NAME}{postfix}_{label}"] = prediction[:, i]
        else:
            df_pred[f"{PREDICTION_COLUMN_NAME}{postfix}"] = prediction

    return df_pred


def __loss(df, model):
    df_loss = pd.DataFrame({}, index=df.index)
    goals = model.features_and_labels.get_goals()
    for target, (loss, _) in goals.items():
        postfix = f"_{target}" if len(goals) > 1 else ""
        if loss in df.columns:
            df_loss[f"{LOSS_COLUMN_NAME}{postfix}_{loss}"] = df[loss]
        else:
            df_loss[f"{LOSS_COLUMN_NAME}{postfix}"] = -abs(loss) if loss is not None else -1.0

    return df_loss


def __truth(df, model):
    df_truth = pd.DataFrame({}, index=df.index)
    goals = model.features_and_labels.get_goals()

    for target, (_, labels) in goals.items():
        postfix = "" if target is None else f'_{target}'
        for label in labels:
            postfix2 = "" if len(labels) <= 1 else f"_{label}"
            df_truth[f"{LABEL_COLUMN_NAME}{postfix}{postfix2}"] = df[label]

    return df_truth


def __stack_header_prediction(goals):
    prediction_headers = []
    # prediction
    for target, (loss, labels) in goals.items():
        prediction_headers.append((target or TARGET_COLUMN_NAME, TARGET_COLUMN_NAME, "value"))
        for l in labels:
            prediction_headers.append((target or TARGET_COLUMN_NAME, PREDICTION_COLUMN_NAME, l if len(labels) > 1 else "value"))

    return prediction_headers


def __stack_header_label(goals):
    return [(target or TARGET_COLUMN_NAME, LABEL_COLUMN_NAME, l if len(labels) > 1 else "value") for target, (_, labels) in goals.items() for l in labels ]


def __stack_header_loss(goals):
    return [(target or TARGET_COLUMN_NAME, LOSS_COLUMN_NAME, "value") for target, (loss, _) in goals.items()]



"""Augment pandas DataFrame with methods for machine learning"""
__version__ = '0.0.13'

from .pandas_utils_extension import *
from .analysis.correlation_analysis import *
from .wrappers.lazy_dataframe import *
from .datafetching.fetch_yahoo import *
from .model.models import *
from .model.features_and_Labels import *
from pandas_ml_utils.analysis.selection import *
from .classification.summary import *
from .classification.classifier import *
from .reinforcement.agent import *
from .regression.regressor import *
from .train_test_data import *
from pandas.core.base import PandasObject


# add functions to pandas
# general utility functions
PandasObject.inner_join = inner_join
PandasObject.drop_re = drop_re
PandasObject.drop_zero_or_nan = drop_zero_or_nan
PandasObject.add_apply = add_apply
PandasObject.shift_inplace = shift_inplace
PandasObject.extend_forecast = extend_forecast
PandasObject.make_training_data = make_training_data

# feature selection
PandasObject.plot_correlation_matrix = plot_correlation_matrix
PandasObject.feature_selection = feature_selection

# classification functions
PandasObject.fit_classifier = fit_classifier
PandasObject.classify = classify
PandasObject.backtest_classifier = backtest_classifier

# regression functions
PandasObject.fit_regressor = fit_regressor
PandasObject.backtest_regressor = backtest_regressor
PandasObject.regress = regress

# reinforcement learning
PandasObject.fit_agent = fit_agent
PandasObject.backtest_agent = backtest_agent
PandasObject.agent_take_action = agent_take_action

# data fetcher
setattr(pd, 'fetch_yahoo', fetch_yahoo)

__doc__ = """
The main concept is to extend pandas DataFrame objects such that you can apply any statistical or machine learning
model directly to the DataFrame.

* feature selection
   - :code:`df.plot_correlation_matrix()`
   - :code:`df.filtration()`

* classification
   - :code:`df.fit_classifier(model)`
   - :code:`df.classify(model)`
   - :code:`df.backtest_classifier(model)`

* regression functions
   - :code:`df.fit_regressor(model)`
   - :code:`df.backtest_regressor(model)`
   - :code:`df.regress(model)`

* reinforcement learning
   - :code:`df.fit_agent(model)`
   - :code:`df.backtest_agent(model)`
   - :code:`df.agent_take_action(model)`
  
Where a model is composed of a ML :class:`.Model` and a :class:`.FeaturesAndLabels` object. Every `fit_` returns a 
:class:`.Fit` which provides a :class:`.Summary` and a :code:`.save_model('./models/super.model')` method. Models can
be loaded back via :code:`Model.load('./models/super.model')`.
"""


import pandas as pd


class HashableDataFrame(object):

    def __init__(self, df: pd.DataFrame) -> None:
        self.df: pd.DataFrame = df

    def __getitem__(self, item: str):
        return self.df.__getitem__(item)

    def __getattr__(self, item):
        return self.df.__getattr__(item)

    def __hash__(self):
        return hash(str(self.describe()))

    def __eq__(self, other):
        try:
            pd.testing.assert_frame_equal(self.df, other.df)
            return True
        except:
            return False



import logging
from collections.abc import Iterable
from itertools import groupby

from typing import Tuple, Dict, Callable, Any, Union, List

import pandas_ml_utils.model.fit
from .utils import unfold_parameter_space
import pandas_ml_utils as pdu
import dill as pickle
import pandas as pd
import numpy as np


log = logging.getLogger(__name__)


# TODO we want multi model to somehow become a Model
class MultiModel(object):

    @staticmethod
    def load(filename: str):
        with open(filename, 'rb') as file:
            multi_model = pickle.load(file)
            if isinstance(multi_model, MultiModel):
                return multi_model
            else:
                raise ValueError("file provided was not a MultiModel")

    def __init__(self,
                 data_provider: Callable[[], pd.DataFrame],
                 data_engineer: Callable[[pd.DataFrame], pd.DataFrame],
                 model_provider: Callable[[], pdu.Model],
                 parameter_space: Dict[str, Iterable]):
        self.data_provider = data_provider
        self.data_engineer = data_engineer
        self.model_provider = model_provider
        self.parameter_space = unfold_parameter_space(parameter_space.copy(), {})
        self.min_needed_data: int = None
        self.data: pd.DataFrame = None
        self.fits: List[pandas_ml_utils.model.fit.Fit] = None
        self._heatmap_cache = None

    def save(self, filename: str):
        with open(filename, 'wb') as file:
            pickle.dump(self, file)

    # def fetch_data_and_fit(self, test_size: float = 0.4, test_validate_split_seed: int=None):
    #     self.fetch_data()
    #     self.fit()

    def fetch_data(self):
        self._heatmap_cache = None
        self.data = self.data_provider()

    def fit(self,
            test_size: float = 0.4,
            test_validate_split_seed: int = None,
            cross_validation: Tuple[int, Callable[[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]] = None
            ) -> None:
        def model_fitter(**kwargs) -> pandas_ml_utils.model.fit.Fit:
            fit = self.data_engineer(self.data, **kwargs) \
                      .fit_classifier(self.model_provider,
                                      test_size=test_size,
                                      cross_validation=cross_validation,
                                      test_validate_split_seed=test_validate_split_seed)

            log.info(f'fit for { {**kwargs}}\n{fit.test_summary.confusion_count()}\n{fit.test_summary.confusion_count()}')
            return fit

        # TODO there should be a way to generate one ClassificationSummary out of several by summing or averaging
        self.fits = [model_fitter(**kwargs) for kwargs in self.parameter_space]
        self.min_needed_data = max([fit.model.min_required_data for fit in self.fits])

    def predict(self) -> pd.DataFrame:
        df = self.data[-self.min_needed_data:] if self.min_needed_data is not None else self.data

        def model_predictor(model, **kwargs) -> pd.DataFrame:
            prediction = self.data_engineer(df, **kwargs) \
                             .classify(model)

            return prediction[-1:]

        predictions = [model_predictor(self.fits[i].model, **kwargs) for i, kwargs in enumerate(self.parameter_space)]
        return predictions

    def plot_heatmap(self, parameter_as_column: str, figsize=(15, 12)):
        import matplotlib.pyplot as plt
        import seaborn as sns

        if self._heatmap_cache is None:
            self._heatmap_cache = self.compute_heatmap(parameter_as_column)

        fig = plt.figure(figsize=figsize)
        sns.heatmap(self._heatmap_cache, annot=True, cmap=plt.cm.Reds)
        return fig

    def compute_heatmap(self, parameter_as_column: str):
        predictions = self.predict()

        # to group all ro indices per column index we first need to sort accordingly
        sorted_parameter_space = sorted(enumerate(self.parameter_space), key=lambda x: x[1][parameter_as_column])

        columns = {col: [value[0] for value in parameter]
                   for col, parameter in groupby(sorted_parameter_space, lambda x: x[1][parameter_as_column])}

        # assign a data frame for each column
        predictions = [pd.concat([predictions[row][["traget_target", "prediction_proba"]] for row in rows], axis=0, sort=True) \
                         .set_index("traget_target") \
                         .groupby(level=0).max() \
                         .rename(columns={"prediction_proba": column})
                       for column, rows in columns.items()]

        predictions = pd.concat(predictions, axis=1, sort=True).sort_index(ascending=False)
        return predictions
